"""Enum Serialization
"""

import unittest

from servicestack import to_json, convert
from servicestack.utils import *
from tests.dtos import HelloWithEnum
from .dtos import *


class TestEnumSerialization(unittest.TestCase):

    def assert_HelloWithEnum(self, actual: HelloWithEnum, expected: HelloWithEnum):
        self.assertEqual(actual.enum_prop, expected.enum_prop)
        self.assertEqual(actual.enum_with_values, expected.enum_with_values)
        self.assertEqual(actual.nullable_enum_prop, expected.nullable_enum_prop)
        self.assertEqual(actual.enum_flags, expected.enum_flags)
        self.assertEqual(actual.enum_style, expected.enum_style)

    def test_does_serialize_HelloWithEnum_empty(self):
        dto = HelloWithEnum()
        json_str = to_json(dto)
        json_obj = json.loads(json_str)
        from_json_obj = convert(HelloWithEnum, json_obj)
        self.assertEqual(from_json_obj, dto)

    def test_does_serialize_HelloWithEnum_EnumFlags(self):
        dto = HelloWithEnum(enum_flags=EnumFlags.VALUE1)
        json_str = to_json(dto)
        json_obj = json.loads(json_str)
        from_json_obj = convert(HelloWithEnum, json_obj)
        self.assertEqual(from_json_obj, dto)

    def test_does_serialize_HelloWithEnum_all(self):
        dto = HelloWithEnum(
            enum_prop=EnumType.VALUE2,
            enum_with_values=EnumWithValues.VALUE1,
            enum_flags=EnumFlags.VALUE1,
            enum_style=EnumStyle.UPPER)
        json_str = to_json(dto)
        json_obj = json.loads(json_str)
        from_json_obj = convert(HelloWithEnum, json_obj)
        self.assertEqual(from_json_obj, dto)

    def assert_HelloWithEnumMap(self, actual: HelloWithEnumMap, expected: HelloWithEnumMap):
        self.assertDictEqual(actual.enum_prop, expected.enum_prop)
        self.assertDictEqual(actual.enum_with_values, expected.enum_with_values)
        self.assertDictEqual(actual.nullable_enum_prop, expected.nullable_enum_prop)
        self.assertDictEqual(actual.enum_flags, expected.enum_flags)
        self.assertDictEqual(actual.enum_style, expected.enum_style)

    def test_does_serialize_HelloWithEnumMap_empty(self):
        dto = HelloWithEnumMap()
        json_str = to_json(dto)
        json_obj = json.loads(json_str)
        from_json_obj = convert(HelloWithEnumMap, json_obj)
        self.assertEqual(from_json_obj, dto)

    def test_does_serialize_HelloWithEnumMap_all(self):
        dto = HelloWithEnumMap(
            enum_prop={f"{EnumType.VALUE2}": EnumType.VALUE2},
            enum_with_values={f"{EnumWithValues.VALUE1}": EnumWithValues.VALUE1},
            enum_flags={f"{EnumFlags.VALUE1}": EnumFlags.VALUE1},
            enum_style={f"{EnumStyle.UPPER}": EnumStyle.UPPER})
        json_str = to_json(dto)
        json_obj = json.loads(json_str)
        from_json_obj = convert(HelloWithEnumMap, json_obj)
        self.assertEqual(from_json_obj, dto)



import base64
import json
import re
from datetime import datetime, timezone, timedelta
from typing import Optional, Any

from .log import Log


def lowercase(string): return str(string).lower()


def uppercase(string): return str(string).upper()


def snakecase(string):
    string = re.sub(r"[\-\.\s]", '_', str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(r"[A-Z]", lambda matched: '_' + lowercase(matched.group(0)), string[1:])


def camelcase(string):
    string = re.sub(r"\w[\s\W]+\w", '', str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(r"[\-_\.\s]([a-z])", lambda matched: uppercase(matched.group(1)), string[1:])


def capitalcase(string: str):
    string = str(string)
    if not string:
        return string
    return uppercase(string[0]) + string[1:]


def pascalcase(string: str): return capitalcase(camelcase(string))


def titlecase(string): return ' '.join([capitalcase(word) for word in snakecase(string).split("_")])


def clean_camelcase(key: str):
    use_key = camelcase(key)
    if use_key[-1] == '_':
        use_key = use_key[0:-1]
    return use_key


def ex_message(e: Exception):
    if hasattr(e, 'message'):
        return e.message
    return str(e)


def log(o: Any):
    print(o)
    return o


def index_of(target: str, needle: str):
    try:
        return target.index(needle)
    except ValueError:
        return -1


def last_index_of(target: str, needle: str):
    try:
        return target.rindex(needle)
    except ValueError:
        return -1


def left_part(str_val: Optional[str], needle: str):
    if str_val is None:
        return None
    pos = index_of(str_val, needle)
    return str_val if pos == -1 else str_val[:pos]


def right_part(str_val: Optional[str], needle: str):
    if str_val is None:
        return None
    pos = index_of(str_val, needle)
    return str_val if pos == -1 else str_val[pos + len(needle):]


def last_left_part(str_val: Optional[str], needle: str):
    if str_val is None:
        return None
    pos = last_index_of(str_val, needle)
    return str_val if pos == -1 else str_val[:pos]


def last_right_part(str_val: Optional[str], needle: str):
    if str_val is None:
        return None
    pos = last_index_of(str_val, needle)
    return str_val if pos == -1 else str_val[pos + len(needle):]


def split_on_first(s: Optional[str], c: str):
    if str is None or str == "":
        return [s]
    pos = index_of(s, c)
    if pos >= 0:
        return [s[:pos], s[pos + 1:]]
    return [s]


def split_on_last(s: Optional[str], c: str):
    if str is None or str == "":
        return [s]
    pos = last_index_of(s, c)
    if pos >= 0:
        return [s[:pos], s[pos + 1:]]
    return [s]


def to_timespan(duration: timedelta):
    total_seconds = duration.total_seconds()
    whole_seconds = total_seconds // 1
    seconds = whole_seconds
    sec = int(seconds % 60 if seconds >= 60 else seconds)
    seconds = seconds // 60
    min = int(seconds % 60)
    seconds = seconds // 60
    hours = int(seconds % 60)
    days = seconds // 24
    remaining_secs = float(sec + (total_seconds - whole_seconds))

    sb = ["P"]
    if days > 0:
        sb.append(f"{days}D")

    if days == 0 or hours + min + sec + remaining_secs > 0:
        sb.append("T")
        if hours > 0:
            sb.append(f"{hours}H")
        if min > 0:
            sb.append(f"{min}M")

        if remaining_secs > 0:
            sec_fmt = "{:.7f}".format(remaining_secs)
            sec_fmt = sec_fmt.rstrip('0')
            sec_fmt = sec_fmt.rstrip('.')
            sb.append(sec_fmt)
            sb.append("S")
        elif len(sb) == 2:  # PT
            sb.append("0S")

    xsd = ''.join(sb)
    # print(f"XSD: {xsd}, {days}:{hours}:{min}:{remaining_secs}")
    return xsd


def from_timespan(s: Optional[str]):
    if s is None:
        return None
    days = 0
    hours = 0
    minutes = 0
    seconds = 0
    ms = 0.0

    if s[0] != "P":
        raise ValueError(f"{s} is not a valid XSD Duration")

    s = s[1:]  # strip P
    t = split_on_first(s, 'T')
    has_time = len(t) == 2

    d = split_on_first(t[0], 'D')
    if len(d) == 2:
        days = int(d[0])

    if has_time:
        h = split_on_first(t[1], 'H')
        if len(h) == 2:
            hours = int(h[0])

        m = split_on_first(h[len(h) - 1], 'M')
        if len(m) == 2:
            minutes = int(m[0])

        s = split_on_first(m[len(m) - 1], 'S')
        if len(s) == 2:
            ms = float(s[0])

        seconds = int(ms)
        ms -= seconds

    # print(f"\n\ntimedelta({str})[{has_time}] = {hours}:{minutes}:{seconds}\n\n")
    return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds, milliseconds=int(ms * 1000))


_MIN_UTC_DATE = datetime.min.replace(tzinfo=timezone.utc)
_MIN_EPOCH = _MIN_UTC_DATE.timestamp()
_MAX_UTC_DATE = datetime.max.replace(tzinfo=timezone.utc)


def to_datetime(date: datetime):
    try:
        return f"/Date({int(date.timestamp() * 1000)})/"
    except Exception as e:
        Log.debug(f"to_datetime({date}): e")
        return None


def from_datetime(json_date: str):
    if json_date.startswith("/Date("):
        epoch_and_zone = left_part(right_part(json_date, "("), ")")
        epoch_str = epoch_and_zone
        if index_of(epoch_and_zone[1:], '-') >= 0:
            epoch_str = last_left_part(epoch_and_zone, '-')
        if index_of(epoch_and_zone[1:], '+') >= 0:
            epoch_str = last_left_part(epoch_and_zone, '+')
        epoch = int(epoch_str)
        try:
            return datetime.fromtimestamp(epoch / 1000, timezone.utc)
        except Exception as e:
            if epoch < _MIN_EPOCH:
                return _MIN_UTC_DATE
            else:
                return _MAX_UTC_DATE

    # need to reduce to 6f precision and remove trailing Z
    has_sec_fraction = index_of(json_date, '.') >= 0
    is_utc = json_date.endswith('Z')
    if is_utc:
        json_date = json_date[0:-1]
    if has_sec_fraction:
        sec_fraction = last_right_part(json_date, '.')
        tz = ''
        if '+' in sec_fraction:
            tz = '+' + right_part(sec_fraction, '+')
            sec_fraction = left_part(sec_fraction, '+')
        elif '-' in sec_fraction:
            sec_fraction = left_part(sec_fraction, '-')
        if len(sec_fraction) > 6:
            json_date = last_left_part(json_date, '.') + '.' + sec_fraction[0:6] + tz

    if is_utc:
        return datetime.fromisoformat(json_date).replace(tzinfo=timezone.utc)
    else:
        return datetime.fromisoformat(json_date)


def to_bytearray(value: Optional[bytes]):
    if value is None:
        return None
    return base64.b64encode(value).decode('ascii')


def from_bytearray(base64str: Optional[str]):
    return base64.b64decode(base64str)


def from_base64url_safe(input_str: str):
    output = input_str
    output = output.replace('-', '+')
    output = output.replace('_', '/')
    pad = len(output) % 4
    if pad == 2:
        output += "=="
    elif pad == 3:
        output += "="
    elif pad != 0:
        raise ValueError("Illegal base46url string!")
    return base64.b64decode(output)


def _decode_base64url_payload(payload: str):
    payload_bytes = from_base64url_safe(payload)
    payload_json = payload_bytes.decode('utf-8')
    return json.loads(payload_json)


def inspect_jwt(jwt: str):
    head = _decode_base64url_payload(left_part(jwt, '.'))
    body = _decode_base64url_payload(left_part(right_part(jwt, '.'), '.'))
    exp = int(body['exp'])
    return head, body, datetime.fromtimestamp(exp, timezone.utc)



import base64
import collections
import decimal
import inspect
import json
import math
import numbers
import os
import pathlib
import platform
import typing
import sys
from dataclasses import fields, is_dataclass
from datetime import datetime, timedelta
from enum import Enum, IntEnum, EnumMeta
from functools import reduce
from types import MappingProxyType
from typing_extensions import Type, get_origin, get_args
from typing import Callable, ForwardRef, Union, TypeVar
from typing import List, Dict, Any

import marshmallow.fields as mf

from servicestack.log import Log
from servicestack.utils import to_timespan, to_datetime, to_bytearray, from_bytearray, from_datetime, from_timespan
from .utils import snakecase, uppercase, camelcase


def is_optional(cls: Type): return get_origin(cls) is Union and type(None) in get_args(cls)


def is_list(cls: Type): return cls == typing.List or cls == list or get_origin(cls) == list


def is_dict(cls: Type): return cls == typing.Dict or cls == dict or get_origin(cls) == dict or cls == MappingProxyType


def nameof(instance):
    return type(instance).__name__


def generic_args(cls: Type):
    if not hasattr(cls, '__args__'):
        raise TypeError(f"{cls} is not a Generic Type")
    return cls.__args__


def generic_arg(cls: Type): return generic_args(cls)[0]


def _get_type_vars_map(cls: Type, type_map=None):
    if type_map is None:
        type_map = {}
    if hasattr(cls, '__orig_bases__'):
        for base_cls in cls.__orig_bases__:
            _get_type_vars_map(base_cls, type_map)
    generic_def = get_origin(cls)
    if generic_def is not None:
        generic_type_args = get_args(cls)
        i = 0
        for t in generic_def.__parameters__:
            k = t
            if type(k) == TypeVar:
                k = k.__name__
            type_map[k] = generic_type_args[i]
            i += 1
    return type_map


def _dict_with_string_keys(d: dict):
    to = {}
    for k, v in d.items():
        to[f"{k}"] = v
    return to


def has_type_vars(cls: Type):
    if cls is None:
        return None
    return isinstance(cls, TypeVar) or any(isinstance(x, TypeVar) for x in cls.__args__)


def _empty(x):
    return x is None or x == {} or x == []


def _str(x: Any):
    if type(x) == str:
        return x
    return f"{x}"


def identity(x: Any): return x


_JSON_TYPES = {str, bool, int, float}
_BUILT_IN_TYPES = {str, bool, int, float, decimal.Decimal, datetime, timedelta, bytes, bytearray, complex}


def is_builtin(t: Type):
    try:
        return t in _BUILT_IN_TYPES or issubclass(t, Enum)
    except Exception as e:  # throws if t is not hashable
        return False


def to_dict(obj: Any, key_case: Callable[[str], str] = identity, remove_empty: bool = True):
    t = type(obj)
    if obj is None or is_builtin(t):
        return obj
    if is_list(t):
        to = []
        for o in obj:
            use_val = to_dict(o, key_case=key_case, remove_empty=remove_empty)
            if not remove_empty or use_val is not None:
                to.append(use_val)
    elif is_dict(t):
        to = {}
        for k, v in obj.items():
            use_key = key_case(_str(k))
            use_val = to_dict(v, key_case=key_case, remove_empty=remove_empty)
            if not remove_empty or use_val is not None:
                to[use_key] = use_val
    elif hasattr(obj, 'to_dict'):  # dataclass
        d = obj.to_dict()
        to = {}
        for k, v in d.items():
            use_key = key_case(_str(k))
            use_val = to_dict(v, key_case=key_case, remove_empty=remove_empty)
            if not remove_empty or use_val is not None:
                to[use_key] = use_val
    elif hasattr(obj, '__dict__'):
        to = to_dict(vars(obj), key_case=key_case, remove_empty=remove_empty)
    else:
        to = obj
    if remove_empty:
        return clean_any(to)
    return to


def _clean_list(d: list):
    return [v for v in (clean_any(v) for v in d) if not _empty(v)]


def _clean_dict(d: dict):
    return {k: v for k, v in ((k, clean_any(v)) for k, v in d.items()) if not _empty(v)}


def clean_any(d):
    """recursively remove empty lists, empty dicts, or None elements from a dictionary"""
    if is_dict(d):
        return _clean_dict(d)
    elif is_list(d):
        return _clean_list(d)
    else:
        return d


def _json_encoder(obj: Any):
    t = type(obj)
    if is_dataclass(t) or is_dict(t):
        return to_dict(obj)
    if isinstance(obj, datetime):
        return obj.isoformat()
    if isinstance(obj, timedelta):
        return to_timespan(obj)
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode('ascii')
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if t in _JSON_TYPES:
        return obj
    if is_builtin(t):
        return _str(obj)
    raise TypeError(f"Unsupported Type in JSON encoding: {t}")


def to_json(obj: Any, indent=None):
    return json.dumps(to_dict(obj), indent=indent, default=_json_encoder)


class TypeConverters:
    serializers = {}    # Py3.9: dict[Type, Callable[[Any], Any]]
    deserializers = {}  # Py3.9: dict[Type, Callable[[Any], Any]]

    @staticmethod
    def register(cls: Type, serializer: Callable[[Any], Any] = None, deserializer: Callable[[Any], Any] = None):
        if serializer is not None:
            TypeConverters.serializers[cls] = serializer
        if deserializer is not None:
            TypeConverters.deserializers[cls] = deserializer

    @staticmethod
    def serialize(obj: Any):
        cls = type(obj)
        if cls in TypeConverters.serializers:
            serializer = TypeConverters.serializers[cls]
            try:
                return serializer(obj)
            except Exception as e:
                Log.error(f"serializer(obj) {cls}({obj})", e)
                raise e

    @staticmethod
    def deserialize(cls: Type, obj: Any):
        deserializer = TypeConverters.deserializers[cls]
        try:
            return deserializer(obj)
        except Exception as e:
            Log.error(f"deserializer(obj) {cls}({obj})", e)
            raise e


TypeConverters.serializers = {
    datetime: to_datetime,
    timedelta: to_timespan,
    bytes: to_bytearray,
    bytearray: to_bytearray,
}
TypeConverters.deserializers = {
    mf.DateTime: from_datetime,
    mf.TimeDelta: from_timespan,
    datetime: from_datetime,
    timedelta: from_timespan,
    bytes: from_bytearray,
    bytearray: from_bytearray,
}


def _resolve_forwardref(cls: Type, orig: Type = None):
    type_name = cls.__forward_arg__
    if orig is not None and orig.__name__ == type_name:
        return orig
    if type_name not in globals():
        raise TypeError(f"Could not resolve ForwardRef('{type_name}')")
    return globals()[type_name]


def unwrap(cls: Type, module:str):
    if type(cls) == ForwardRef:
        cls = _resolve_forwardref(cls)
    if isinstance(cls, str):
        cls = eval(cls, {}, vars(sys.modules[module]))
    if is_optional(cls):
        return generic_arg(cls)
    return cls


def dict_get(name: str, obj: dict, case: Callable[[str], str] = None):
    if name in obj:
        return obj[name]
    if case:
        name_case = case(name)
        if name_case in obj:
            return obj[name_case]
    name_snake = snakecase(name)
    if name_snake in obj:
        return obj[name_snake]
    name_camel = camelcase(name)
    if name_camel in obj:
        return obj[name_camel]
    if name.endswith('_'):
        return dict_get(name.rstrip('_'), obj, case)
    return None


def sanitize_name(s: str):
    return s.replace('_', '').upper()


def enum_get(cls: Union[Enum, Type], key: Union[str, int]):
    if type(key) == int or issubclass(cls, IntEnum):
        return cls(key)
    try:
        return cls[key]
    except Exception as e:
        try:
            upper_snake = uppercase(snakecase(key))
            return cls[upper_snake]
        except Exception as e2:
            sanitize_key = sanitize_name(key)
            for value in cls.__members__.values():
                if sanitize_key == sanitize_name(value):
                    return value
            for member in cls.__members__.keys():
                if sanitize_key == sanitize_name(member):
                    return cls[member]
    raise TypeError(f"{key} is not a member of {nameof(Enum)}")


def _resolve_type(cls: Type, substitute_types: Dict[Type, type] ):
    if type(cls) == TypeVar:
        # TypeVar('T') == TypeVar('T') is false, 
        # I think this should work
        cls = cls.__name__
    if substitute_types is None:
        return cls
    return substitute_types[cls] if cls in substitute_types else cls


def convert(into: Type, obj: Any, substitute_types: Dict[Type, type] = None, module = None):
    if obj is None:
        return None
    into = unwrap(into, module)
    into = _resolve_type(into, substitute_types)
    if Log.debug_enabled():
        Log.debug(f"convert({into}, {substitute_types}, {obj})")

    is_type = type(into) == type
    if not is_type:
        Log.debug(f"type of {into} is not a class")

    generic_def = get_origin(into)
    if generic_def is not None and is_dataclass(generic_def):
        reified_types = _get_type_vars_map(into)
        return convert(generic_def, obj, reified_types)

    if is_dataclass(into):
        to = {}
        for f in fields(into):
            val = dict_get(f.name, obj)
            # print(f"to[{f.name}] = convert({f.type}, {val}, {substitute_types})")
            to[f.name] = convert(f.type, val, substitute_types, into.__module__)
            # print(f"to[{f.name}] = {to[f.name]}")
        return into(**to)
    elif is_list(into):
        el_type = _resolve_type(generic_arg(into), substitute_types)
        to = []
        for item in obj:
            to.append(convert(el_type, item, substitute_types, into.__module__))
        return to
    elif is_dict(into):
        key_type, val_type = generic_args(into)
        key_type = _resolve_type(key_type, substitute_types)
        val_type = _resolve_type(val_type, substitute_types)
        to = {}
        if not hasattr(obj, 'items'):
            Log.warn(f"dict {obj} ({type(type)}) does not have items()")
        for key, val in obj.items():
            to_key = convert(key_type, key, substitute_types, into.__module__)
            to_val = convert(val_type, val, substitute_types, into.__module__)
            to[to_key] = to_val
        return to
    else:
        if into in TypeConverters.deserializers:
            converter = TypeConverters.deserializers[into]
            try:
                return converter(obj)
            except Exception as e:
                Log.error(f"converter(obj) {into}({obj})", e)
                raise e
        elif inspect.isclass(into) and issubclass(into, mf.Field):
            try:
                return into().deserialize(obj)
            except Exception as e:
                Log.error(f"into().deserialize(obj) {into}({obj})", e)
                raise e
        elif is_type and (issubclass(into, Enum) or into == EnumMeta):
            try:
                return enum_get(into, obj)
            except Exception as e:
                print(into, type(into), obj, type(obj))
                Log.error(f"Enum into[obj] {into}[{obj}]", e)
                raise e
        else:
            try:
                return into(obj)
            except Exception as e:
                # if into == typing.Dict or get_origin(into) == typing.Dict:
                #     print("WAS A typing.Dict")
                Log.error(f"into(obj) {into}({obj})", e)
                raise e


def from_json(into: Type, json_str: str):
    if json_str is None or json_str == "":
        return None
    json_obj = json.loads(json_str)
    return convert(into, json_obj)


# inspect utils
def all_keys(obj):
    keys = []
    if not isinstance(obj, collections.Iterable):
        return keys
    for o in obj:
        if is_builtin(type(o)):
            continue
        for key in o:
            key = _str(key)
            if key is not None and key not in keys:
                keys.append(key)
    return keys


def inspect_vars(objs):
    if not isinstance(objs, dict):
        raise TypeError('objs must be a dictionary')

    to = to_dict(objs)

    inspect_vars_path = os.environ.get('INSPECT_VARS')
    if inspect_vars_path is None:
        return
    if platform.system() == 'Windows':
        inspect_vars_path = inspect_vars_path.replace("/", "\\")
    else:
        inspect_vars_path = inspect_vars_path.replace("\\", "/")

    pathlib.Path(os.path.dirname(inspect_vars_path)).mkdir(parents=True, exist_ok=True)

    with open(inspect_vars_path, 'w') as outfile:
        json_str = to_json(to)
        outfile.write(json_str)


def dump(obj):
    return to_json(obj, indent=4).replace('"', '').replace(': null', ':')


def printdump(obj):
    print(dump(obj))


def _align_left(s: str, length: int, pad: str = ' '):
    if length < 0:
        return ""
    alen = length + 1 - len(s)
    if alen <= 0:
        return s
    return pad + s + (pad * (length + 1 - len(s)))


def _align_center(s: str, length: int, pad: str = ' '):
    if length < 0:
        return ""
    nlen = len(s)
    half = math.floor(length / 2 - nlen / 2)
    odds = abs((nlen % 2) - (length % 2))
    return (pad * (half + 1)) + s + (pad * (half + 1 + odds))


def _align_right(s: str, length: int, pad: str = ' '):
    if length < 0:
        return ""
    alen = length + 1 - len(s)
    if alen <= 0:
        return s
    return (pad * (length + 1 - len(s))) + s + pad


def _align_auto(obj: Any, length: int, pad: str = ' '):
    s = _str(obj)
    if len(s) <= length:
        if isinstance(obj, numbers.Number):
            return _align_right(s, length, pad)
        return _align_left(s, length, pad)
    return s


def sanitize_key(key: str): return key.replace("_", " ").replace(" ", "").lower()


def lenient_getitem(obj: dict, key: str):
    if key in obj:
        return obj[key]
    sanitized_key = sanitize_key(key)
    for k in obj:
        if sanitize_key(k) == sanitized_key:
            return obj[k]
    return ""


def table(objs, headers=None):
    if not is_list(type(objs)):
        raise TypeError('objs must be a list')
    map_rows = to_dict(objs)
    if headers is None:
        headers = all_keys(map_rows)
    col_sizes: Dict[str, int] = {}

    for k in headers:
        max = len(k)
        for row in map_rows:
            col = lenient_getitem(row, k)
            if col != "":
                val_size = len(_str(col))
                if val_size > max:
                    max = val_size
            col_sizes[k] = max

    # sum + ' padding ' + |
    row_width = reduce(lambda x, y: x + y, col_sizes.values(), 0) + \
                (len(col_sizes) * 2) + \
                (len(col_sizes) + 1)
    sb: List[str] = [f"+{'-' * (row_width - 2)}+"]
    head = "|"
    for k in headers:
        head += _align_center(k, col_sizes[k]) + "|"
    sb.append(head)
    sb.append(f"|{'-' * (row_width - 2)}|")

    for row in map_rows:
        to = "|"
        for k in headers:
            to += '' + _align_auto(lenient_getitem(row, k), col_sizes[k]) + "|"
        sb.append(to)

    sb.append(f"+{'-' * (row_width - 2)}+")
    return '\n'.join(sb)


def printtable(obj, headers=None):
    print(table(obj, headers))


def htmllist(d: dict):
    sb: List[str] = ["<table><tbody>"]
    for k, v in d.items():
        sb.append(f"<tr><th>{_str(k)}</th><td>{htmldump(v)}</td></tr>")
    sb.append("</tbody></table>")
    return ''.join(sb)


def htmldump(objs, headers=None):
    if is_builtin(type(objs)):
        return _str(objs)

    map_rows = to_dict(objs)
    t = type(map_rows)
    if is_dict(t):
        return htmllist(map_rows)

    if headers is None:
        headers = all_keys(map_rows)

    # print(headers, map_rows)

    sb: List[str] = ["<table>"]
    row = ["<thead><tr>"]
    for k in headers:
        row.append(f"<th>{k}</th>")
    row.append("</tr></head>")
    if len(row) > 2:
        sb.append(''.join(row))
    sb.append("<tbody>")

    rows = []
    for item in map_rows:
        rows.append("<tr>")
        if len(headers) > 0:
            row = []
            for k in headers:
                val = lenient_getitem(item, k)
                row.append(f"<td>{htmldump(val)}</td>")
            rows.append(''.join(row))
        else:
            rows.append(f"<td>{htmldump(item)}</td>")
        rows.append("</tr>")

    sb.append(''.join(rows))
    sb.append("</tbody>")
    sb.append("</table>")
    return '\n'.join(sb)


def printhtmldump(obj, headers=None):
    print(htmldump(obj, headers))



"""Serialization Tests
"""
import unittest

from servicestack.clients import append_querystring
from servicestack.utils import *
from tests.config import *
from .dtos import *
from tests.test_client import create_AllTypes, create_AllCollectionTypes, \
    assert_AllTypes, assert_AllCollectionTypes

JSON_ALL_TYPES = '{"id":1,"byte":2,"short":3,"int":4,"long":5,"uShort":6,"uInt":7,"uLong":8,"float":1.1,"double":2.2,"decimal":3.0,"string":"string","dateTime":"\/Date(978307200000)\/","timeSpan":"PT1H","dateTimeOffset":"\/Date(978307200000)\/","guid":"ea762009b66c410b9bf5ce21ad519249","char":"c","keyValuePair":{},"stringList":["A","B","C"],"stringArray":["D","E","F"],"stringMap":{"A":"D","B":"E","C":"F"},"intStringMap":{"1":"A","2":"B","3":"C"},"subType":{"id":1,"name":"name"}}'
JSON_ALL_COLLECTION_TYPES = '{"intArray":[1,2,3],"intList":[4,5,6],"stringArray":["A","B","C"],"stringList":["D","E","F"],byteArray:"QUJD","pocoArray":[{"name":"pocoArray"}],"pocoList":[{"name":"pocoList"}],"pocoLookup":{"A":[{"name":"B"},{"name":"C"}]},"pocoLookupMap":{"A":[{"B":{"name":"C"}},{"D":{"name":"E"}}]}}'


class TestSerialization(unittest.TestCase):

    def test_does_serialize_AllTypes(self):
        dto = create_AllTypes()
        json_str = to_json(dto)
        json_obj = from_json(AllTypes, json_str)
        assert_AllTypes(self, json_obj)

    def test_does_deserialize_AllTypes(self):
        json_obj = from_json(AllTypes, JSON_ALL_TYPES)
        assert_AllTypes(self, json_obj)

    def test_does_serialize_AllCollectionTypes(self):
        dto = create_AllCollectionTypes()
        json_str = to_json(dto)
        json_obj = from_json(AllCollectionTypes, json_str)
        assert_AllCollectionTypes(self, json_obj)

    def test_does_serialize_HelloAllTypes(self):
        dto = HelloAllTypes(
            name="HelloAllTypes",
            all_types=create_AllTypes(),
            all_collection_types=create_AllCollectionTypes())
        json_str = to_json(dto)
        json_obj: HelloAllTypes = from_json(HelloAllTypes, json_str)
        self.assertEqual(dto.name, json_obj.name)
        assert_AllTypes(self, json_obj.all_types)
        assert_AllCollectionTypes(self, json_obj.all_collection_types)

    def test_does_serialize_inheritance(self):
        dto = HelloWithInheritance(name="foo", id=1)
        json_str = to_json(dto)
        json_obj: HelloWithInheritance = from_json(HelloWithInheritance, json_str)
        self.assertEqual(json_obj.name, "foo")
        self.assertEqual(json_obj.id, 1)



from servicestack import JsonServiceClient, Authenticate

# TEST_URL = "https://localhost:5001"
# TEST_URL = "http://localhost:5000"
TEST_URL = "http://test.servicestack.net"
TECHSTACKS_URL = "https://techstacks.io"


def create_test_client():
    return JsonServiceClient(TEST_URL)


def create_techstacks_client():
    return JsonServiceClient(TECHSTACKS_URL)


def clear_session(client: JsonServiceClient):
    client.post(Authenticate(provider="logout"))



__all__ = [
    'IReturn',
    'IReturnVoid',
    'IGet',
    'IPost',
    'IPut',
    'IDelete',
    'IPatch',
    'IOptions',
    'IMeta',
    'IHasSessionId',
    'IHasBearerToken',
    'IHasVersion',
    'ICrud',
    'ICreateDb',
    'IUpdateDb',
    'IPatchDb',
    'IDeleteDb',
    'ISaveDb',
    'KeyValuePair',
    'ResponseError',
    'ResponseStatus',
    'QueryBase',
    'QueryDb',
    'QueryDb1',
    'QueryDb2',
    'QueryData',
    'QueryData2',
    'QueryResponse',
    'AuthenticateResponse',
    'Authenticate',
    'AssignRolesResponse',
    'AssignRoles',
    'UnAssignRolesResponse',
    'UnAssignRoles',
    'ConvertSessionToTokenResponse',
    'ConvertSessionToToken',
    'GetAccessTokenResponse',
    'GetAccessToken',
    'CancelRequestResponse',
    'CancelRequest',
    'UpdateEventSubscriber',
    'UpdateEventSubscriberResponse',
    'UserApiKey',
    'GetApiKeys',
    'GetApiKeysResponse',
    'RegenerateApiKeys',
    'RegenerateApiKeysResponse',
    'NavItem',
    'GetNavItems',
    'GetNavItemsResponse',
    'EmptyResponse',
    'IdResponse',
    'StringResponse',
    'StringsResponse',
    'AuditBase',
    'BackgroundJobState',
    'JobStatSummary',
    'HourSummary',
    'WorkerStats',
    'BackgroundJobBase',
    'BackgroundJob',
    'JobSummary',
    'BackgroundJobOptions',
    'ScheduledTask',
    'CompletedJob',
    'FailedJob',
    'JsonServiceClient',
    'WebServiceException',
    'WebServiceExceptionType',
    'qsvalue',
    'resolve_httpmethod',
    'TypeConverters',
    'to_json',
    'from_json',
    'all_keys',
    'to_dict',
    'convert',
    'is_optional',
    'is_list',
    'is_dict',
    'is_builtin',
    'generic_args',
    'generic_arg',
    'inspect_vars',
    'dump',
    'printdump',
    'table',
    'printtable',
    'htmldump',
    'printhtmldump',
    'lowercase',
    'uppercase',
    'snakecase',
    'camelcase',
    'capitalcase',
    'pascalcase',
    'titlecase',
    'clean_camelcase',
    'index_of',
    'last_index_of',
    'left_part',
    'right_part',
    'last_left_part',
    'last_right_part',
    'split_on_first',
    'split_on_last',
    'to_timespan',
    'from_timespan',
    'from_datetime',
    'to_bytearray',
    'from_bytearray',
    'from_base64url_safe',
    'inspect_jwt',
    'Bytes',
]

from .dtos import \
    IReturn, IReturnVoid, IGet, IPost, IPut, IDelete, IPatch, IOptions, IMeta, \
    IHasSessionId, IHasBearerToken, IHasVersion, ICrud, \
    ICreateDb, IUpdateDb, IPatchDb, IDeleteDb, ISaveDb, \
    KeyValuePair, ResponseError, ResponseStatus, QueryBase, \
    QueryDb, QueryDb1, QueryDb2, QueryData, QueryData2, QueryResponse, \
    AuthenticateResponse, Authenticate, AssignRolesResponse, AssignRoles, \
    UnAssignRolesResponse, UnAssignRoles, ConvertSessionToTokenResponse, \
    ConvertSessionToToken, GetAccessTokenResponse, GetAccessToken, \
    CancelRequestResponse, CancelRequest, UpdateEventSubscriber, \
    UpdateEventSubscriberResponse, UserApiKey, GetApiKeys, GetApiKeysResponse, \
    RegenerateApiKeys, RegenerateApiKeysResponse, NavItem, GetNavItems, \
    GetNavItemsResponse, EmptyResponse, IdResponse, StringResponse, \
    StringsResponse, AuditBase, \
    BackgroundJobState, JobStatSummary, HourSummary, WorkerStats, BackgroundJobBase, \
    BackgroundJob, JobSummary, BackgroundJobOptions, ScheduledTask, CompletedJob, FailedJob


from .clients import JsonServiceClient, WebServiceException, WebServiceExceptionType, \
    qsvalue, resolve_httpmethod

from .reflection import TypeConverters, to_json, from_json, all_keys, to_dict, convert, is_optional, \
    is_list, is_dict, is_builtin, generic_args, generic_arg, inspect_vars, dump, printdump, table, \
    printtable, htmldump, printhtmldump

from .utils import \
    lowercase, uppercase, snakecase, camelcase, capitalcase, pascalcase, titlecase, clean_camelcase, \
    index_of, last_index_of, left_part, right_part, last_left_part, last_right_part, \
    split_on_first, split_on_last, to_timespan, from_timespan, from_datetime, \
    to_bytearray, from_bytearray, from_base64url_safe, inspect_jwt

from .fields import Bytes



from typing import TypeVar, Generic, Optional, Dict, List, Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined
from datetime import datetime, timedelta
from enum import Enum, IntEnum

T = TypeVar('T')
Table = TypeVar('Table')
From = TypeVar('From')
Into = TypeVar('Into')
TKey = TypeVar('TKey')
TValue = TypeVar('TValue')


class IReturn(Generic[T]):
    pass


class IReturnVoid:
    pass


class IGet:
    pass


class IPost:
    pass


class IPut:
    pass


class IDelete:
    pass


class IPatch:
    pass


class IOptions:
    pass


class IMeta:
    meta: Optional[Dict[str, str]] = None


class IHasSessionId:
    session_id: Optional[str] = None


class IHasBearerToken:
    bearer_token: Optional[str] = None


class IHasVersion:
    version: Optional[int] = None


class ICrud:
    pass


class ICreateDb(Generic[Table]):
    pass


class IUpdateDb(Generic[Table]):
    pass


class IPatchDb(Generic[Table]):
    pass


class IDeleteDb(Generic[Table]):
    pass


class ISaveDb(Generic[Table]):
    pass


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class KeyValuePair(Generic[TKey, TValue]):
    key: TKey = None
    value: TValue = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ResponseError:
    error_code: Optional[str] = None
    field_name: Optional[str] = None
    message: Optional[str] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ResponseStatus:
    error_code: Optional[str] = None
    message: Optional[str] = None
    stack_trace: Optional[str] = None
    errors: List[ResponseError] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryBase:
    skip: Optional[int] = None
    take: Optional[int] = None
    order_by: Optional[str] = None
    order_by_desc: Optional[str] = None
    include: Optional[str] = None
    fields: Optional[str] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryResponse(Generic[T]):
    offset: int = None
    total: int = None
    results: List[T] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryDb(Generic[T], QueryBase, IReturn[QueryResponse[T]]):
    @staticmethod
    def response_type(): return QueryResponse[T]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryDb1(Generic[T], QueryBase, IReturn[QueryResponse[T]]):
    @staticmethod
    def response_type(): return QueryResponse[T]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryDb2(Generic[From, Into], QueryBase, IReturn[QueryResponse[Into]]):
    @staticmethod
    def response_type(): return QueryResponse[Into]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryData(Generic[T], QueryBase, IReturn[QueryResponse[T]]):
    @staticmethod
    def response_type(): return QueryResponse[T]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryData1(Generic[From, Into], QueryBase, IReturn[QueryResponse[Into]]):
    @staticmethod
    def response_type(): return QueryResponse[Into]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class QueryData2(Generic[From, Into], QueryBase, IReturn[QueryResponse[Into]]):
    @staticmethod
    def response_type(): return QueryResponse[Into]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AuthenticateResponse(IHasSessionId, IHasBearerToken):
    user_id: Optional[str] = None
    session_id: Optional[str] = None
    user_name: Optional[str] = None
    display_name: Optional[str] = None
    referrer_url: Optional[str] = None
    bearer_token: Optional[str] = None
    refresh_token: Optional[str] = None
    profile_url: Optional[str] = None
    roles: Optional[List[str]] = None
    permissions: Optional[List[str]] = None
    response_status: Optional[ResponseStatus] = None
    meta: Optional[Dict[str, str]] = None


# @Route("/auth")
# @Route("/auth/{provider}")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class Authenticate(IReturn[AuthenticateResponse], IPost):
    provider: Optional[str] = None
    state: Optional[str] = None
    oauth_token: Optional[str] = None
    oauth_verifier: Optional[str] = None
    user_name: Optional[str] = None
    password: Optional[str] = None
    remember_me: Optional[bool] = None
    error_view: Optional[str] = None
    nonce: Optional[str] = None
    uri: Optional[str] = None
    response: Optional[str] = None
    qop: Optional[str] = None
    nc: Optional[str] = None
    cnonce: Optional[str] = None
    use_token_cookie: Optional[bool] = None
    access_token: Optional[str] = None
    access_token_secret: Optional[str] = None
    scope: Optional[str] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AssignRolesResponse:
    all_roles: Optional[List[str]] = None
    all_permissions: Optional[List[str]] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


# @Route("/assignroles")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AssignRoles(IReturn[AssignRolesResponse], IPost):
    user_name: Optional[str] = None
    permissions: Optional[List[str]] = None
    roles: Optional[List[str]] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UnAssignRolesResponse:
    all_roles: Optional[List[str]] = None
    all_permissions: Optional[List[str]] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


# @Route("/unassignroles")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UnAssignRoles(IReturn[UnAssignRolesResponse], IPost):
    user_name: Optional[str] = None
    permissions: Optional[List[str]] = None
    roles: Optional[List[str]] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ConvertSessionToTokenResponse:
    meta: Optional[Dict[str, str]] = None
    access_token: Optional[str] = None
    refresh_token: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


# @Route("/session-to-token")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ConvertSessionToToken(IReturn[ConvertSessionToTokenResponse], IPost):
    preserve_session: bool = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetAccessTokenResponse:
    access_token: Optional[str] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


# @Route("/access-token")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetAccessToken(IReturn[GetAccessTokenResponse], IPost):
    refresh_token: Optional[str] = None
    use_token_cookie: Optional[bool] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CancelRequestResponse:
    tag: Optional[str] = None
    elapsed: timedelta = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CancelRequest(IPost):
    tag: Optional[str] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateEventSubscriber(IPost):
    id: Optional[str] = None
    subscribe_channels: Optional[List[str]] = None
    unsubscribe_channels: Optional[List[str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UpdateEventSubscriberResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class UserApiKey:
    key: Optional[str] = None
    key_type: Optional[str] = None
    expiry_date: Optional[datetime] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetApiKeys(IGet):
    environment: Optional[str] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetApiKeysResponse:
    results: List[UserApiKey] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RegenerateApiKeysResponse:
    results: List[UserApiKey] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class RegenerateApiKeys(IPost):
    environment: Optional[str] = None
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class NavItem:
    label: Optional[str] = None
    href: Optional[str] = None
    exact: Optional[bool] = None
    id: Optional[str] = None
    class_name: Optional[str] = None
    icon_class: Optional[str] = None
    show: Optional[str] = None
    hide: Optional[str] = None
    children: List[Any] = None  # NavItem
    meta: Optional[Dict[str, str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetNavItems:
    name: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetNavItemsResponse:
    base_url: Optional[str] = None
    results: List[NavItem] = None
    nav_items_map: Dict[str, List[NavItem]] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class EmptyResponse:
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class IdResponse:
    id: Optional[str] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class StringResponse:
    results: Optional[str] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class StringsResponse:
    results: Optional[List[str]] = None
    meta: Optional[Dict[str, str]] = None
    response_status: Optional[ResponseStatus] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class AuditBase:
    created_date: datetime = None
    # @Required()
    created_by: Optional[str] = None

    modified_date: datetime = None
    # @Required()
    modified_by: Optional[str] = None

    deleted_date: Optional[datetime] = None
    deleted_by: Optional[str] = None


class BackgroundJobState(str, Enum):
    QUEUED = 'Queued'
    STARTED = 'Started'
    EXECUTED = 'Executed'
    COMPLETED = 'Completed'
    FAILED = 'Failed'
    CANCELLED = 'Cancelled'

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class JobStatSummary:
    name: Optional[str] = None
    total: int = 0
    completed: int = 0
    retries: int = 0
    failed: int = 0
    cancelled: int = 0

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class HourSummary:
    hour: Optional[str] = None
    total: int = 0
    completed: int = 0
    failed: int = 0
    cancelled: int = 0

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class WorkerStats:
    name: Optional[str] = None
    queued: int = 0
    received: int = 0
    completed: int = 0
    retries: int = 0
    failed: int = 0
    running_job: Optional[int] = None
    running_time: Optional[timedelta] = None

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class BackgroundJobBase:
    id: int = 0
    parent_id: Optional[int] = None
    ref_id: Optional[str] = None
    worker: Optional[str] = None
    tag: Optional[str] = None
    batch_id: Optional[str] = None
    callback: Optional[str] = None
    depends_on: Optional[int] = None
    run_after: Optional[datetime] = None
    created_date: datetime = datetime(1, 1, 1)
    created_by: Optional[str] = None
    request_id: Optional[str] = None
    request_type: Optional[str] = None
    command: Optional[str] = None
    request: Optional[str] = None
    request_body: Optional[str] = None
    user_id: Optional[str] = None
    response: Optional[str] = None
    response_body: Optional[str] = None
    state: Optional[BackgroundJobState] = None
    started_date: Optional[datetime] = None
    completed_date: Optional[datetime] = None
    notified_date: Optional[datetime] = None
    retry_limit: Optional[int] = None
    attempts: int = 0
    duration_ms: int = 0
    timeout_secs: Optional[int] = None
    progress: Optional[float] = None
    status: Optional[str] = None
    logs: Optional[str] = None
    last_activity_date: Optional[datetime] = None
    reply_to: Optional[str] = None
    error_code: Optional[str] = None
    error: Optional[ResponseStatus] = None
    args: Optional[Dict[str, str]] = None
    meta: Optional[Dict[str, str]] = None

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class BackgroundJob(BackgroundJobBase):
    id: int = 0

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class JobSummary:
    id: int = 0
    parent_id: Optional[int] = None
    ref_id: Optional[str] = None
    worker: Optional[str] = None
    tag: Optional[str] = None
    batch_id: Optional[str] = None
    created_date: datetime = datetime(1, 1, 1)
    created_by: Optional[str] = None
    request_type: Optional[str] = None
    command: Optional[str] = None
    request: Optional[str] = None
    response: Optional[str] = None
    user_id: Optional[str] = None
    callback: Optional[str] = None
    started_date: Optional[datetime] = None
    completed_date: Optional[datetime] = None
    state: Optional[BackgroundJobState] = None
    duration_ms: int = 0
    attempts: int = 0
    error_code: Optional[str] = None
    error_message: Optional[str] = None

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class BackgroundJobOptions:
    ref_id: Optional[str] = None
    parent_id: Optional[int] = None
    worker: Optional[str] = None
    run_after: Optional[datetime] = None
    callback: Optional[str] = None
    depends_on: Optional[int] = None
    user_id: Optional[str] = None
    retry_limit: Optional[int] = None
    reply_to: Optional[str] = None
    tag: Optional[str] = None
    batch_id: Optional[str] = None
    created_by: Optional[str] = None
    timeout_secs: Optional[int] = None
    timeout: Optional[timedelta] = None
    args: Optional[Dict[str, str]] = None
    run_command: Optional[bool] = None

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ScheduledTask:
    id: int = 0
    name: Optional[str] = None
    interval: Optional[timedelta] = None
    cron_expression: Optional[str] = None
    request_type: Optional[str] = None
    command: Optional[str] = None
    request: Optional[str] = None
    request_body: Optional[str] = None
    options: Optional[BackgroundJobOptions] = None
    last_run: Optional[datetime] = None
    last_job_id: Optional[int] = None

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class CompletedJob(BackgroundJobBase):
    pass

@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class FailedJob(BackgroundJobBase):
    pass



""" Basic Serialization Tests
"""

import unittest
import dataclasses
import json
from .dtos import *
from datetime import datetime, timedelta, timezone
from servicestack import JsonServiceClient, WebServiceException, to_json
from .config import create_test_client



def create_HelloAllTypes():
    return HelloAllTypes(name="name",
                         all_types=create_AllTypes(),
                         all_collection_types=create_AllCollectionTypes())


def create_AllTypes():
    return AllTypes(
        id=1,
        char='c',
        byte=2,
        short=3,
        int_=4,
        long=5,
        u_short=6,
        u_int=7,
        u_long=8,
        float_=1.1,
        double=2.2,
        decimal=decimal.Decimal(3.0),
        string="string",
        date_time=datetime(2001, 1, 1, tzinfo=timezone.utc),
        date_time_offset=datetime(2001, 1, 1, tzinfo=timezone.utc),
        time_span=timedelta(hours=1),
        guid="ea762009b66c410b9bf5ce21ad519249",
        string_list=["A", "B", "C"],
        string_array=["D", "E", "F"],
        string_map={"A": "D", "B": "E", "C": "F"},
        int_string_map={1: "A", 2: "B", 3: "C"},
        sub_type=SubType(id=1, name="name"))


def create_AllCollectionTypes():
    return AllCollectionTypes(
        int_array=[1, 2, 3],
        int_list=[1, 2, 3],
        string_array=["A", "B", "C"],
        string_list=["D", "E", "F"],
        byte_array=b"ABC",  # base64(ABC)
        poco_array=[create_Poco("pocoArray")],
        poco_list=[create_Poco("pocoArray")],
        poco_lookup={"A": [create_Poco("B"), create_Poco("C")]},
        poco_lookup_map={"A": [{"B": create_Poco("C"), "D": create_Poco("E")}]})


def create_Poco(name: str): return Poco(name=name)


def create_EchoComplexTypes():
    return EchoComplexTypes(
        sub_type=SubType(id=1, name="foo"),
        sub_types=[SubType(id=2, name="bar"), SubType(id=3, name="baz")],
        sub_type_map={
            "a": SubType(id=4, name="qux")
        },
        string_map={
            "a": "b"
        },
        int_string_map={
            1: "A"
        }
    )


def assert_AllTypes(test: unittest.TestCase, dto: AllTypes):
    # print(type(dto))
    # print(vars(dto))
    test.assertEqual(dto.id, 1)
    test.assertEqual(dto.byte, 2)
    test.assertEqual(dto.short, 3)
    test.assertEqual(dto.int_, 4)
    test.assertEqual(dto.long, 5)
    test.assertEqual(dto.u_short, 6)
    test.assertEqual(dto.u_int, 7)
    test.assertEqual(dto.u_long, 8)
    test.assertEqual(dto.float_, 1.1)
    test.assertEqual(dto.double, 2.2)
    test.assertEqual(dto.decimal, 3.0)
    test.assertEqual(dto.string, "string")
    test.assertEqual(dto.date_time, datetime(2001, 1, 1, tzinfo=timezone.utc))
    test.assertEqual(dto.date_time_offset, datetime(2001, 1, 1, tzinfo=timezone.utc))
    test.assertEqual(dto.time_span, timedelta(hours=1))
    test.assertEqual(dto.guid, "ea762009b66c410b9bf5ce21ad519249")
    test.assertListEqual(dto.string_list, ["A", "B", "C"])
    test.assertListEqual(dto.string_array, ["D", "E", "F"])
    test.assertDictEqual(dto.string_map, {"A": "D", "B": "E", "C": "F"})
    test.assertDictEqual(dto.int_string_map, {1: "A", 2: "B", 3: "C"})
    test.assertEqual(dto.sub_type.id, 1)
    test.assertEqual(dto.sub_type.name, "name")


def assert_AllCollectionTypes(test: unittest.TestCase, dto: AllCollectionTypes):
    test.assertListEqual(dto.int_array, [1, 2, 3])
    test.assertListEqual(dto.int_list, [1, 2, 3])
    test.assertListEqual(dto.string_array, ["A", "B", "C"])
    test.assertListEqual(dto.string_list, ["D", "E", "F"])
    test.assertEqual(dto.byte_array, b'ABC')
    test.assertEqual(len(dto.poco_array), 1)
    test.assertEqual(dto.poco_array[0].name, "pocoArray")
    test.assertEqual(len(dto.poco_lookup), 1)
    poco_lookup_values = dto.poco_lookup["A"]
    test.assertEqual(len(poco_lookup_values), 2)
    test.assertEqual(poco_lookup_values[0].name, "B")
    test.assertEqual(poco_lookup_values[1].name, "C")
    test.assertEqual(len(dto.poco_lookup_map), 1)
    poco_lookup_map_values = dto.poco_lookup_map["A"]
    test.assertEqual(len(poco_lookup_map_values), 1)
    poco_lookup_mapa_list = poco_lookup_map_values[0]
    test.assertEqual(len(poco_lookup_mapa_list), 2)
    test.assertEqual(poco_lookup_mapa_list["B"].name, "C")
    test.assertEqual(poco_lookup_mapa_list["D"].name, "E")


client = create_test_client()


class TestApi(unittest.TestCase):

    def assert_HelloAllTypesResponse(self, dto: HelloAllTypesResponse):
        # print(dto)
        self.assertEqual(dto.result, "name")
        self.assert_AllTypes(dto.all_types)
        self.assert_AllCollectionTypes(dto.all_collection_types)

    def assert_AllTypes(self, dto: AllTypes): assert_AllTypes(self, dto)

    def assert_AllCollectionTypes(self, dto: AllCollectionTypes):
        assert_AllCollectionTypes(self, dto)

    def assert_EchoComplexTypes(self, dto: EchoComplexTypes):
        self.assertEqual(dto.sub_type.id, 1)
        self.assertEqual(dto.sub_type.name, "foo")
        self.assertEqual(dto.sub_types[0].id, 2)
        self.assertEqual(dto.sub_types[0].name, "bar")
        self.assertEqual(dto.sub_types[1].id, 3)
        self.assertEqual(dto.sub_types[1].name, "baz")
        self.assertEqual(dto.sub_type_map["a"].id, 4)
        self.assertEqual(dto.sub_type_map["a"].name, "qux")
        self.assertEqual(dto.string_map["a"], "b")
        self.assertEqual(dto.int_string_map[1], "A")

    def test_can_get_hello(self):
        response: HelloResponse = client.get(Hello(name="World"))
        self.assertEqual(response.result, "Hello, World!")

    def test_can_post_hello(self):
        response: HelloResponse = client.post(Hello(name="World"))
        self.assertEqual(response.result, "Hello, World!")

    def test_can_send_umlauts(self):
        response: HelloResponse = client.post(Hello(name="üöäß"))
        self.assertEqual(response.result, "Hello, üöäß!")

    def test_does_fire_Request_and_Response_filters(self):
        client = create_test_client()
        events = []

        JsonServiceClient.global_request_filter = lambda info: events.append("globalRequestFilter")
        JsonServiceClient.global_response_filter = lambda res: events.append("globalResponseFilter")

        client.request_filter = lambda info: events.append("requestFilter")
        client.response_filter = lambda res: events.append("responseFilter")

        response: HelloResponse = client.get(Hello(name="World"))
        self.assertEqual(response.result, "Hello, World!")

        self.assertListEqual(events, [
            "requestFilter",
            "globalRequestFilter",
            "responseFilter",
            "globalResponseFilter"
        ])

        JsonServiceClient.global_request_filter = None
        JsonServiceClient.global_response_filter = None

    def test_can_get_hello_with_custom_path(self):
        response: HelloResponse = client.get_url("/hello/World", response_as=HelloResponse)
        self.assertEqual(response.result, "Hello, World!")

    def test_can_get_hello_with_CustomPath_as_raw_types(self):
        json_str = client.get_url("/hello", response_as=str, args={'name': 'World'})
        self.assertEqual(json_str, '{"result":"Hello, World!"}')

        json_bytes: bytes = client.get_url("/hello", response_as=bytes, args={'name': 'World'})
        self.assertEqual(json_bytes.decode("utf-8"), '{"result":"Hello, World!"}')

        dto: HelloResponse = client.get_url("/hello", response_as=HelloResponse, args={'name': 'World'})
        self.assertEqual(dto.result, "Hello, World!")

    def test_can_post_Hello_with_CustomPath(self):
        response: HelloResponse = client.post_url("/hello", Hello(name="World"))
        self.assertEqual(response.result, "Hello, World!")

    def test_can_post_hello_with_CustomPath_json_object(self):
        json_obj = client.post_url("/hello", to_json(Hello(name="World")))
        self.assertIsInstance(json_obj, dict)
        response = HelloResponse(**json_obj)
        self.assertEqual(response.result, "Hello, World!")

    def test_can_post_HelloAllTypes(self):
        request = create_HelloAllTypes()
        response: HelloAllTypesResponse = client.post(request)
        self.assert_HelloAllTypesResponse(response)

    def test_can_put_HelloAllTypes(self):
        request = create_HelloAllTypes()
        response: HelloAllTypesResponse = client.put(request)
        self.assert_HelloAllTypesResponse(response)

    def test_does_handle_404_error(self):
        request = ThrowType(type="NotFound", message="not here")
        try:
            client.put(request)
            self.fail("should throw")
        except WebServiceException as ex:
            status = ex.response_status
            self.assertEqual(status.error_code, "NotFound")
            self.assertEqual(status.message, "not here")

    def test_does_handle_ValidationException(self):
        request = ThrowValidation(email="invalidemail")
        try:
            client.post(request)
            self.fail("should throw")
        except WebServiceException as ex:
            status = ex.response_status
            errors = status.errors
            self.assertEqual(len(errors), 3)
            self.assertEqual(errors[0].error_code, status.error_code)
            self.assertEqual(errors[0].message, status.message)

            self.assertEqual(errors[0].error_code, "InclusiveBetween")
            self.assertEqual(errors[0].message, "'Age' must be between 1 and 120. You entered 0.")
            self.assertEqual(errors[0].field_name, "Age")

            self.assertEqual(errors[1].error_code, "NotEmpty")
            self.assertEqual(errors[1].message, "'Required' must not be empty.")
            self.assertEqual(errors[1].field_name, "Required")

            self.assertEqual(errors[2].error_code, "Email")
            self.assertEqual(errors[2].message, "'Email' is not a valid email address.")
            self.assertEqual(errors[2].field_name, "Email")

    def test_does_handle_auth_failure(self):
        request = RequiresAdmin()
        try:
            client.post(request)
            self.fail("should throw")
        except WebServiceException as ex:
            self.assertEqual(ex.status_code, 401)

    def test_can_send_ReturnVoid(self):
        client = create_test_client()
        sent_methods = []
        client.request_filter = lambda req: sent_methods.append(req.method)

        request = SendReturnVoid(id=1)
        client.send(request)
        self.assertEqual(sent_methods[-1], "POST")
        request.id = 2
        client.get(request)
        self.assertEqual(sent_methods[-1], "GET")
        request.id = 3
        client.post(request)
        self.assertEqual(sent_methods[-1], "POST")
        request.id = 4
        client.put(request)
        self.assertEqual(sent_methods[-1], "PUT")
        request.id = 5
        client.delete(request)
        self.assertEqual(sent_methods[-1], "DELETE")

    def test_can_get_response_as_Raw_String(self):
        response = client.get(HelloString(name="World"))
        self.assertEqual(response, "World")

    def test_can_get_response_as_Raw_Bytes(self):
        response = client.get_url("/json/reply/HelloString?Name=World", response_as=bytes)
        self.assertEqual(response.decode("utf-8"), "World")

    def test_should_return_raw_text(self):
        response = client.get(ReturnString(data="0x10"))
        self.assertEqual(response, "0x10")

    def test_can_send_raw_json_as_object(self):
        client = create_test_client()
        client.response_filter = lambda res: self.assertEqual(res.headers["X-Args"], "1,name")

        body = {"foo": "bar"}
        request = SendJson(id=1, name="name")

        json_str = client.post(request, body=to_json(body))
        json_obj = json.loads(json_str)

        self.assertEqual(json_obj["foo"], "bar")

    def test_can_send_raw_string(self):
        client = create_test_client()
        client.response_filter = lambda res: self.assertEqual(res.headers["X-Args"], "1,name")
        body = "foo"
        request = SendText(id=1, name="name", content_type="text/plain")
        str = client.post(request, body=body)
        self.assertEqual(str, "foo")

    def test_can_deserialize_nested_list(self):
        client = create_test_client()
        response: Items = client.get(GetItems())
        self.assertEqual(len(response.results), 2)

        all_names = list(map(lambda x: x.name, response.results))
        self.assertLessEqual(all_names, ["bar item 1", "bar item 2"])

    def test_can_deserialize_naked_list(self):
        client = create_test_client()
        response: List[Item] = client.get(GetNakedItems())
        self.assertEqual(len(response), 2)
        all_names = list(map(lambda x: x.name, response))
        self.assertLessEqual(all_names, ["item 1", "item 2"])

    def test_can_deserialize_custom_generic_response_type(self):
        response: QueryResponseAlt[Item] = client.get(AltQueryItems())
        self.assertEqual(len(response.results), 2)
        all_names = list(map(lambda x: x.name, response.results))
        self.assertLessEqual(all_names, ["item 1", "item 2"])

    # requires Python 3.9
    def test_can_send_all_batch_request(self):
        client = create_test_client()
        client.response_filter = lambda res: self.assertEqual(res.headers["X-AutoBatch-Completed"], "3")
        requests = list(map(lambda name: Hello(name=name), ["foo", "bar", "baz"]))
        responses = client.send_all(requests)
        self.assertListEqual(list(map(lambda x: x.result, responses)),
                             ['Hello, foo!', 'Hello, bar!', 'Hello, baz!'])

    # requires Python 3.9
    def test_can_send_all_oneway_IReturn_batch_request(self):
        client = create_test_client()
        client.request_filter = lambda req: self.assertTrue(req.url.endswith("/json/oneway/Hello[]"))
        requests = list(map(lambda name: Hello(name=name), ["foo", "bar", "baz"]))
        client.send_all_oneway(requests)

    # requires Python 3.9
    def test_can_send_all_oneway_IReturnVoid_batch_request(self):
        client = create_test_client()
        client.request_filter = lambda req: self.assertTrue(req.url.endswith("/json/oneway/HelloReturnVoid[]"))
        requests = list(map(lambda name: HelloReturnVoid(name=name), [1, 2, 3]))
        client.send_all_oneway(requests)

    def test_can_post_to_EchoTypes(self):
        response: EchoTypes = client.post(EchoTypes(int_=1, string="foo"))
        self.assertEqual(response.int_, 1)
        self.assertEqual(response.string, "foo")

    def test_can_get_IReturnVoid_requests(self):
        client.get(HelloReturnVoid(id=1))

    def test_can_post_IReturnVoid_requests(self):
        client.post(HelloReturnVoid(id=1))

    def test_can_handle_Validation_Errors_with_camelcasing(self):
        client = create_test_client()
        client.request_filter = lambda req: self.assertTrue(
            req.url.endswith("ThrowValidation?jsconfig=EmitCamelCaseNames%3Atrue"))
        try:
            client.post(ThrowValidation(), args={"jsconfig": "EmitCamelCaseNames:true"})
        except WebServiceException as e:
            self.assertEqual(e.response_status.error_code, "InclusiveBetween")
            self.assertEqual(e.response_status.message, "'Age' must be between 1 and 120. You entered 0.")
            self.assertEqual(e.response_status.errors[1].error_code, "NotEmpty")
            self.assertEqual(e.response_status.errors[1].field_name, "Required")
            self.assertEqual(e.response_status.errors[1].message, "'Required' must not be empty.")

    def test_can_handle_Validation_Errors_with_pascalcasing(self):
        client = create_test_client()
        client.request_filter = lambda req: self.assertTrue(
            req.url.endswith("ThrowValidation?jsconfig=EmitCamelCaseNames%3Afalse"))
        try:
            client.post(ThrowValidation(), args={"jsconfig": "EmitCamelCaseNames:false"})
        except WebServiceException as e:
            self.assertEqual(e.response_status.error_code, "InclusiveBetween")
            self.assertEqual(e.response_status.message, "'Age' must be between 1 and 120. You entered 0.")
            self.assertEqual(e.response_status.errors[1].error_code, "NotEmpty")
            self.assertEqual(e.response_status.errors[1].field_name, "Required")
            self.assertEqual(e.response_status.errors[1].message, "'Required' must not be empty.")

    def test_can_get_using_only_path_info(self):
        response: HelloResponse = client.get_url("/hello/World", response_as=HelloResponse)
        self.assertEqual(response.result, "Hello, World!")

    def test_can_get_using_absolute_url(self):
        response: HelloResponse = client.get_url("http://test.servicestack.net/hello/World", response_as=HelloResponse)
        self.assertEqual(response.result, "Hello, World!")

    def test_can_get_using_route_and_querystring(self):
        response: HelloResponse = client.get_url("/hello", args={"name": "World"}, response_as=HelloResponse)
        self.assertEqual(response.result, "Hello, World!")

    def test_can_get_EchoTypes_using_route(self):
        request = EchoTypes(long=1, string="foo")
        args = dataclasses.asdict(request)
        response: EchoTypes = client.get_url("/echo/types", args=args, response_as=EchoTypes)
        self.assertEqual(response.long, 1)
        self.assertEqual(response.string, "foo")

    def test_can_post_EchoComplexTypes(self):
        request = create_EchoComplexTypes()
        response = client.post(request)
        self.assert_EchoComplexTypes(response)

    def test_can_handle_connection_error(self):
        client = JsonServiceClient("http://unknown-zzz.net")

        client.exception_filter = lambda res, e: (
            # print(e.status_code) and
            # print(e.status_description)
        )
        try:
            client.get(EchoTypes(int_=1, string="foo"))
        except WebServiceException as e:
            self.assertEqual(e.status_code, 500)
            self.assertIn("getaddrinfo failed", e.status_description)
            # self.assertTrue("getaddrinfo failed" in handled_ex.status_description)

    def test_can_handle_naked_List(self):
        request = HelloList(names=['A', 'B', 'C'])
        response: List[ListResult] = client.get(request)
        self.assertEqual(len(response), 3)



"""JsonServiceClient Tests
"""
import unittest

from servicestack.clients import append_querystring
from servicestack.utils import *
from tests.config import *
from .dtos import *


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class NonExistingService(IReturn[str]):
    data: Optional[str] = None


class TestJsonServiceClient(unittest.TestCase):

    def test_does_process_missing_service_correctly(self):
        client = create_test_client()
        try:
            client.get(NonExistingService())
        except WebServiceException as e:
            self.assertEqual(e.status_code, 405)

    def test_does_serialize_dates_correctly_via_get_request(self):
        client = create_test_client()
        request = EchoTypes(date_time=datetime.datetime(2015, 1, 1, tzinfo=timezone.utc))
        response: EchoTypes = client.get(request)
        self.assertEqual(response.date_time, request.date_time)

    def test_should_generate_default_value(self):
        client = create_test_client()
        request = HelloTypes(bool_=False, int_=0)
        request_url = append_querystring(TEST_URL, to_dict(request, key_case=clean_camelcase))
        self.assertEqual(request_url, TEST_URL + "?bool=false&int=0")



from abc import ABC, abstractmethod
from enum import Enum
from typing import List


class LogLevel(str, Enum):
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARN = 'WARN'
    ERROR = 'ERROR'


class Logger(ABC):
    @abstractmethod
    def log(self, level: LogLevel, msg: str, e: Exception = None):
        ...


class ConsoleLogger(Logger):
    def log(self, level: LogLevel, msg: str, e: Exception = None):
        print(f"{level}: {msg}")
        if e:
            print(e)


class NullLogger(Logger):
    def log(self, level: LogLevel, msg: str, e: Exception = None):
        pass


class Log:
    levels: List[LogLevel] = [LogLevel.WARN, LogLevel.ERROR]
    logger: Logger = ConsoleLogger()

    @staticmethod
    def debug_enabled():
        return LogLevel.DEBUG in Log.levels

    @staticmethod
    def info_enabled():
        return LogLevel.INFO in Log.levels

    @staticmethod
    def warn_enabled():
        return LogLevel.WARN in Log.levels

    @staticmethod
    def error_enabled():
        return LogLevel.ERROR in Log.levels

    @staticmethod
    def debug(msg: str):
        if Log.debug_enabled():
            Log.logger.log(LogLevel.DEBUG, msg)

    @staticmethod
    def info(msg: str):
        if Log.info_enabled():
            Log.logger.log(LogLevel.INFO, msg)

    @staticmethod
    def warn(msg: str, e: Exception = None):
        if Log.warn_enabled():
            Log.logger.log(LogLevel.WARN, msg, e)

    @staticmethod
    def error(msg: str, e: Exception = None):
        if Log.error_enabled():
            Log.logger.log(LogLevel.ERROR, msg, e)


# Log.levels.append(LogLevel.DEBUG)



""" Auth Tests
"""

import unittest

from servicestack import WebServiceException
from .config import *
from .dtos import *


def create_jwt(**args) -> CreateJwt:
    to = CreateJwt(**args)
    if to.user_auth_id is None:
        to.user_auth_id = "1"
    if to.display_name is None:
        to.display_name = "test jwt"
    if to.email is None:
        to.email = "test@auth.com"
    return to


class State:
    count: int = 0

    def incr(self):
        self.count += 1


class TestAuthClient(unittest.TestCase):

    def test_can_auth_with_JWT(self):
        client = create_test_client()
        response: CreateJwtResponse = client.post(create_jwt())
        client.bearer_token = response.token

        test_auth: TestAuthResponse = client.get(TestAuth())
        self.assertEqual(test_auth.user_id, "1")
        self.assertEqual(test_auth.display_name, "test jwt")
        self.assertIsNotNone(test_auth.session_id)

    def test_does_fire_on_authentication_required_callback_on_401(self):
        client = create_test_client()
        state = State()

        client.on_authentication_required = lambda s=state: (s.incr())

        try:
            client.get(TestAuth())
            self.fail("should throw")
        except WebServiceException as e:
            status = e.response_status
            self.assertEqual(status.error_code, "401")
            self.assertEqual(status.message, "Unauthorized")
            self.assertEqual(state.count, 1)

    def test_can_use_on_authentication_required_to_auth_client(self):
        client = create_test_client()
        state = State()

        client.on_authentication_required = lambda c=client, s=state: [
            s.incr(),
            c.set_credentials("test", "test")
        ]

        client.get(TestAuth())
        self.assertEqual(state.count, 1)

    def test_can_use_on_authentication_required_to_fetch_token(self):
        client = create_test_client()
        state = State()
        auth_client = create_test_client()

        client.on_authentication_required = lambda c=client, a=auth_client, s=state: [
            s.incr(),
            a.set_credentials("test", "test"),
            client.set_bearer_token(cast(AuthenticateResponse, a.get(Authenticate())).bearer_token)
        ]

        client.get(TestAuth())
        self.assertEqual(state.count, 1)

    def test_can_use_on_authentication_required_to_fetch_token_after_expired_token(self):
        client = create_test_client()
        state = State()

        client.on_authentication_required = lambda c=client, s=state: [
            s.incr(),
            c.set_bearer_token(cast(CreateJwtResponse, c.post(create_jwt())).token)
        ]

        create_expired_jwt = create_jwt()
        create_expired_jwt.jwt_expiry = datetime.datetime(2000, 1, 1)
        expired_jwt: CreateJwtResponse = client.post(create_expired_jwt)

        client.bearer_token = expired_jwt.token
        client.get(TestAuth())
        self.assertEqual(state.count, 1)

    def test_can_use_refresh_token_to_fetch_token_after_expired_token(self):
        client = create_test_client()

        client.set_credentials("test", "test")
        auth_response: AuthenticateResponse = client.post(Authenticate())

        client.refresh_token = auth_response.refresh_token
        client.set_credentials(None, None)

        create_expired_jwt = create_jwt()
        create_expired_jwt.jwt_expiry = datetime.datetime(2000, 1, 1)
        expired_jwt: CreateJwtResponse = client.post(create_expired_jwt)

        client.bearer_token = expired_jwt.token
        client.get(TestAuth())
        self.assertNotEqual(client.bearer_token, expired_jwt.token)

    def test_can_reauthenticate_after_an_auto_refresh_access_token(self):
        client = create_test_client()

        auth = Authenticate(provider="credentials", user_name="test", password="test")
        auth_response: AuthenticateResponse = client.post(auth)

        refresh_token = auth_response.refresh_token

        create_expired_jwt = create_jwt()
        create_expired_jwt.jwt_expiry = datetime.datetime(2000, 1, 1)
        expired_jwt: CreateJwtResponse = client.post(create_expired_jwt)
        bearer_token = expired_jwt.token

        clear_session(client)

        client = create_test_client()
        client.bearer_token = bearer_token
        client.refresh_token = refresh_token

        auth.password = "notvalid"
        try:
            client.post(auth)
            self.fail("should throw")
        except WebServiceException as e:
            status = e.response_status
            self.assertEqual(status.error_code, "Unauthorized")
            self.assertEqual(status.message, "Invalid Username or Password")

    def test_does_fetch_access_token_using_refresh_token_cookies(self):
        client = create_test_client()
        auth = Authenticate(provider="credentials", user_name="test", password="test")
        client.post(auth)

        initial_access_token = client.token_cookie
        initial_refresh_token = client.refresh_token_cookie
        self.assertIsNotNone(initial_access_token)
        self.assertIsNotNone(initial_refresh_token)

        request = Secured(name="test")
        response: SecuredResponse = client.send(request)
        self.assertEqual(response.result, request.name)

        client.post(InvalidateLastAccessToken())

        response: SecuredResponse = client.send(request)
        print(response)
        self.assertEqual(response.result, request.name)

        latest_access_token = client.token_cookie
        self.assertNotEqual(latest_access_token, initial_access_token)

    def test_invalid_refresh_token_throws_RefreshTokenException_ErrorResponse(self):
        client = create_test_client()

        client.refresh_token = "Invalid.Refresh.Token"

        try:
            client.get(TestAuth())
        except WebServiceException as e:
            self.assertEqual(e.type, WebServiceExceptionType.REFRESH_TOKEN_EXCEPTION)
            self.assertEqual(e.response_status.error_code, "ArgumentException")
            self.assertEqual(e.response_status.message, "Illegal base64url string!")

    def test_expires_refresh_token_throws_RefreshTokenException(self):
        client = create_test_client()

        create_expired_jwt = create_jwt()
        create_expired_jwt.jwt_expiry = datetime.datetime(2000, 1, 1)
        expired_jwt: CreateJwtResponse = client.post(create_expired_jwt)
        client.refresh_token = expired_jwt.token

        try:
            client.get(TestAuth())
        except WebServiceException as e:
            self.assertEqual(e.type, WebServiceExceptionType.REFRESH_TOKEN_EXCEPTION)
            self.assertEqual(e.response_status.error_code, "TokenException")
            self.assertEqual(e.response_status.message, "Token has expired")




""" Test Client Utils
"""

import unittest
import re
from .dtos import *
from servicestack import to_json, qsvalue, resolve_httpmethod


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetLocationsResponse:
    locations: Optional[List[str]] = None


# @Route("/locations")
@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class GetLocations(IReturn[GetLocationsResponse], IGet):
    pass


def sanitize_html(s: str): return re.sub(r"\s", "", s)


class ClientUtils(unittest.TestCase):

    def test_json_encode(self):
        h = Hello(name='A')
        self.assertEqual(to_json(None), 'null')
        self.assertEqual(to_json("A"), '"A"')
        self.assertEqual(to_json(1), '1')
        self.assertEqual(to_json(h), '{"name": "A"}')
        self.assertEqual(to_json([h, h]), '[{"name": "A"}, {"name": "A"}]')
        self.assertEqual(to_json({'a': h, 'b': h}), '{"a": {"name": "A"}, "b": {"name": "A"}}')

    def test_qsvalue(self):
        self.assertEqual(qsvalue(None), "")
        self.assertEqual(qsvalue("A"), "A")
        self.assertEqual(qsvalue(1), "1")
        self.assertEqual(qsvalue(b"Python"), "UHl0aG9u")

    def test_does_resolve_IVerbs_from_request_DTO_interface_marker(self):
        self.assertEqual(resolve_httpmethod(SendGet()), "GET")
        self.assertEqual(resolve_httpmethod(SendPost()), "POST")
        self.assertEqual(resolve_httpmethod(SendPut()), "PUT")
        self.assertEqual(resolve_httpmethod(GetLocations()), "GET")

    def test_can_htmldump_different_types(self):
        self.assertEqual(htmldump(1), "1")
        self.assertEqual(htmldump("A"), "A")
        self.assertEqual(htmldump({"A": 1, "B": 2}),
                         "<table><tbody><tr><th>A</th><td>1</td></tr><tr><th>B</th><td>2</td></tr></tbody></table>")
        self.assertEqual(sanitize_html(htmldump(["A", 1])),
                         "<table><tbody><tr><td>A</td></tr><tr><td>1</td></tr></tbody></table>")
        self.assertEqual(sanitize_html(htmldump([{"A": 1, "B": 2}, {"A": 3, "B": 4}])),
                         "<table><thead><tr><th>A</th><th>B</th></tr></head><tbody><tr><td>1</td><td>2</td></tr><tr><td>3</td><td>4</td></tr></tbody></table>")

    def test_can_send_GetLocations(self):
        client = JsonServiceClient("https://covid-vac-watch.netcore.io")
        response = client.send(GetLocations())
        printdump(response)
        printhtmldump(response.locations)
